import numpy as np
import pandas as pd
from tqdm import tqdm
import argparse
import os

# Visualization
import matplotlib
matplotlib.use('agg')
from matplotlib import pyplot as plt


# fixed parameters
cyana_shift_cols = np.array([1, 2, 3])
tab_shift_cols = np.array([9, 8, 7])
cyana_ass_cols = np.array(cyana_shift_cols) + 9
cyana_colnames = np.linspace(1, 15, 15)
heavy_atom_CS_reduction = 20
plot_window = 4


# calculator of CSP
def eucl_dist(a, b):
    # scaling
    diff = a - b
    diff[0] = diff[0] / heavy_atom_CS_reduction
    return np.sqrt(np.sum(diff ** 2))


# transfer of assignment type (CYANA to PIPE)
def process_ass(ass):
    ass1 = ass[0].split('.')[1] + '.' + ass[0].split('.')[0]
    ass2 = ass[1].split('.')[1] + '.' + ass[1].split('.')[0]
    ass3 = ass[2].split('.')[1] + '.' + ass[2].split('.')[0]
    return ass3 + ';' + ass2 + ';' + ass1


# transfer of assignments
def transfer_assignment(tab_file_path,
                        cyana_file_path,
                        flag,
                        CSP_trashold,
                        repeat,
                        w1_shift,
                        w2_shift,
                        w3_shift):

    # get data path
    DataPath = os.path.dirname(tab_file_path)

    # import data
    W1shift = w1_shift
    W2shift = w2_shift
    W3shift = w3_shift
    for repeat_id in range(repeat):
        assignments = pd.read_csv(cyana_file_path,
                                  names=cyana_colnames,
                                  skiprows=6,
                                  delim_whitespace=True,
                                  header=None).to_numpy()
        peaks = pd.read_csv(tab_file_path,
                            skiprows=19,
                            delim_whitespace=True,
                            header=None).to_numpy()

        # shift data
        assignments[:, cyana_shift_cols[0]] -= W1shift
        assignments[:, cyana_shift_cols[1]] -= W2shift
        assignments[:, cyana_shift_cols[2]] -= W3shift

        # histogram of dimensions
        plt.hist([assignments[:, cyana_shift_cols[0]], peaks[:, tab_shift_cols[0]]],
                 bins=100,
                 label=['CYANA', 'PIPE'])
        plt.legend()
        plt.savefig(DataPath + '/' + flag + '_W1_overlay.png')
        plt.close()

        plt.hist([assignments[:, cyana_shift_cols[1]], peaks[:, tab_shift_cols[1]]],
                 bins=100,
                 label=['CYANA', 'PIPE'])
        plt.legend()
        plt.savefig(DataPath + '/' + flag +'_W2_overlay.png')
        plt.close()

        plt.hist([assignments[:, cyana_shift_cols[2]], peaks[:, tab_shift_cols[2]]],
                 bins=100,
                 label=['CYANA', 'PIPE'])
        plt.legend()
        plt.savefig(DataPath + '/' + flag +'_W3_overlay.png')
        plt.close()

        # transfer assignments
        CSP_W1_container = []
        CSP_W2_container = []
        CSP_W3_container = []
        CSP_container = []
        print('STEP', repeat_id+1)
        for assignment_id in tqdm(range(assignments.shape[0])):
            CSVass = assignments[assignment_id, cyana_shift_cols]
            minCSP = 100
            for peak_id in range(peaks.shape[0]):
                CSVpeak = peaks[peak_id, tab_shift_cols]
                CSP = eucl_dist(CSVass, CSVpeak)
                if CSP < minCSP:
                    peak_id_mem = peak_id
                    minCSP = CSP
                    CSP_W1 = CSVass[0] - CSVpeak[0]
                    CSP_W2 = CSVass[1] - CSVpeak[1]
                    CSP_W3 = CSVass[2] - CSVpeak[2]
            CSP_container.append(minCSP)
            CSP_W1_container.append(CSP_W1)
            CSP_W2_container.append(CSP_W2)
            CSP_W3_container.append(CSP_W3)
            if repeat_id == repeat - 1 and minCSP < CSP_trashold:
                if assignments[assignment_id, cyana_ass_cols][0] != '-' and \
                   assignments[assignment_id, cyana_ass_cols][1] != '-' and \
                   assignments[assignment_id, cyana_ass_cols][2] != '-':
                    peaks[peak_id_mem, 30] = process_ass(assignments[assignment_id, cyana_ass_cols])

        # histogram of total CSPs
        plt.hist(np.array(CSP_container),
                 bins=100,
                 range=(0, CSP_trashold * plot_window))
        plt.savefig(DataPath + '/' + flag + '_CSP_step_' + str(repeat_id+1) + '.png')
        plt.close()

        # histogram of CSP along heavy spin
        plt.hist(np.array(CSP_W1_container),
                 bins=100,
                 range=(-CSP_trashold * heavy_atom_CS_reduction * plot_window,
                        CSP_trashold * heavy_atom_CS_reduction * plot_window))
        plt.savefig(DataPath + '/' + flag + '_CSP_W1_step_' + str(repeat_id+1) + '.png')
        plt.close()

        # histogram of CSP along direct dimension
        plt.hist(np.array(CSP_W2_container),
                 bins=100,
                 range=(-CSP_trashold * plot_window,
                        CSP_trashold * plot_window))
        plt.savefig(DataPath + '/' + flag + '_CSP_W2_step_' + str(repeat_id+1) + '.png')
        plt.close()

        # histogram of CSP along indirect dimension
        plt.hist(np.array(CSP_W3_container),
                 bins=100,
                 range=(-CSP_trashold * plot_window,
                        CSP_trashold * plot_window))
        plt.savefig(DataPath + '/' + flag + '_CSP_W3_step_' + str(repeat_id+1) + '.png')
        plt.close()

        # output the values
        n_transfer = len([csp for csp in CSP_container if csp < CSP_trashold])
        print('Number of trasfered peaks:',
              n_transfer,
              'out of',
              len(CSP_container),
              100 * n_transfer / len(CSP_container), '%')

        print('Average W1 shift:', np.mean([i for i in CSP_W1_container if
                                            -CSP_trashold * heavy_atom_CS_reduction * plot_window < i < CSP_trashold * heavy_atom_CS_reduction * plot_window]))
        print('Average W2 shift:', np.mean([i for i in CSP_W2_container if
                                            -CSP_trashold * plot_window < i < CSP_trashold * plot_window]))
        print('Average W3 shift:', np.mean([i for i in CSP_W3_container if
                                            -CSP_trashold * plot_window < i < CSP_trashold * plot_window]))

        W1shift += np.mean([i for i in CSP_W1_container if
                            -CSP_trashold * heavy_atom_CS_reduction * plot_window < i < CSP_trashold * heavy_atom_CS_reduction * plot_window])
        W2shift += np.mean([i for i in CSP_W2_container if
                                            -CSP_trashold * plot_window < i < CSP_trashold * plot_window])
        W3shift += np.mean([i for i in CSP_W3_container if
                                            -CSP_trashold * plot_window < i < CSP_trashold * plot_window])

    # write results
    ass_tab_file_path = tab_file_path[:-4] + '_asn.tab'
    with open(ass_tab_file_path, 'w') as outfile:
        with open(tab_file_path, 'r') as infile:
            lines = infile.readlines()
            for i in range(19):
                outfile.write(lines[i])
            ass_ids = [i for i in range(peaks.shape[0]) if peaks[i, 30] != 'None']
            for i in ass_ids:
                bufer = lines[i + 19]
                outfile.write(bufer.replace('None', str(peaks[i, 30])))

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Correlation extraction from multistate protein bundles')
    parser.add_argument('tab_file_path', type=str,
                        help='tab file path with peaked peaks')
    parser.add_argument('cyana_file_path', type=str,
                        help='cyana file path with assignments')
    parser.add_argument('flag', type=str,
                        help='NOESY regions (N, C_@ALI or C_@ARO)')
    parser.add_argument('--CSP_trashold', type=float,
                        default=0.05,
                        help='CSP thrashold for assignment transfer')
    parser.add_argument('--repeat', type=int,
                        default=5,
                        help='Number of assignment transfer iterations')
    parser.add_argument('--w1_shift', type=float,
                        default=0,
                        help='Number of assignment transfer iterations')
    parser.add_argument('--w2_shift', type=float,
                        default=0,
                        help='Number of assignment transfer iterations')
    parser.add_argument('--w3_shift', type=float,
                        default=0,
                        help='Number of assignment transfer iterations')
    args = parser.parse_args()
    transfer_assignment(args.tab_file_path,
                        args.cyana_file_path,
                        args.flag,
                        args.CSP_trashold,
                        args.repeat,
                        args.w1_shift,
                        args.w2_shift,
                        args.w3_shift)
